/*
 * ICFP 2002 Programming Contest
 * Entry by Seth Bruder
 *
 */

package sethbruder.icfp2002.entry.player;

import java.util.LinkedList;
import java.util.Iterator;

import sethbruder.icfp2002.entry.*;

final class DistMap {

	private static final byte INF_COST = (byte)0xFF;
	private static final byte NO_DIR = 0;
	static final byte NORTH = 1;
	static final byte SOUTH = 2;
	static final byte EAST = 3;
	static final byte WEST = 4;
	private static final byte MIN_DIR = 1;
	private static final byte MAX_DIR = 4;

	private final int maxDist;
	private final int myX;
	private final int myY;

	private final int[][] distMap;
	private final byte[][] dirMap;

	public DistMap(final Facts facts, boolean otherPlayersAreUnpassable) {
		this(facts, facts.myState.x, facts.myState.y,
			 otherPlayersAreUnpassable);
	}

	public DistMap(final Facts facts, final int myX, final int myY,
				   boolean otherPlayersAreUnpassable) {

		final byte[][] board = facts.board.board;
		final int width = board.length;
		final int height = board[0].length;

		this.maxDist = facts.board.getMaxDist();
		this.myX = myX;
		this.myY = myY;

		this.distMap = new int[width][height];
		this.dirMap = new byte[width][height];
		final boolean[][] passableMap = new boolean[width][height];

		PlayerState[] playerStates
			= new PlayerState[facts.otherPlayerStates.size()];
		facts.otherPlayerStates.values().toArray(playerStates);

		computePassableMap(passableMap, board, playerStates,
						   otherPlayersAreUnpassable);
		computeDistMap(passableMap);
	}

	public int getDist(final int destX, final int destY) {
		return distMap[destX][destY];
	}

	public int getFirstMoveType(final int destX, final int destY) {

		int x, y;
		byte dir, prevDir;

		x = destX;
		y = destY;
		prevDir = NO_DIR;
		dir = dirMap[x][y];

		while (dir != NO_DIR) {
			switch(dir) { // back track to previous point in path
				case NORTH:
					y--;
					break;
				case SOUTH:
					y++;
					break;
				case EAST:
					x--;
					break;
				case WEST:
					x++;
					break;
				default:
					throw new RuntimeException();
			}
			prevDir = dir;
			dir = dirMap[x][y];
		}

		if ((x != myX) || (y != myY)
			|| (prevDir == NO_DIR)) return Move.NULL_MOVE_TYPE;
		
		switch(prevDir) { // back track to previous point in path
			case NORTH:
				return Move.NORTH;
			case SOUTH:
				return Move.SOUTH;
			case EAST:
				return Move.EAST;
			case WEST:
				return Move.WEST;
			default:
				throw new RuntimeException("BadDir: " + dir);
		}
	}

	private void computePassableMap(final boolean[][] passableMap, 
									final byte[][] board,
									final PlayerState[] playerStates,
									boolean otherPlayersAreUnpassable) {
		
		final int width = board.length;
		final int height = board[0].length;
		int x, y, i;

		for (x = 0; x < width; x++)
			for (y = 0; y < height; y++) {
				switch (board[x][y]) {
					case Board.EMPTY:
					case Board.BASE:
						passableMap[x][y] = true;
						break;
					case Board.WATER:
					case Board.WALL:
						passableMap[x][y] = false;
						break;
					default:
						throw new RuntimeException();
				}
			}

		if (otherPlayersAreUnpassable) {
			for (i = 0; i < playerStates.length; i++) {
				passableMap[playerStates[i].x][playerStates[i].y] = false;
			}
		}
	}

	private void computeDistMap(final boolean[][] passableMap) {

		final int width = passableMap.length;
		final int height = passableMap[0].length;

		LinkedList frontierNodeList;

		int x, y;
		byte dir;

		for (x = 0; x < width; x++)
			for (y = 0; y < height; y++) {
				distMap[x][y] = maxDist;
				dirMap[x][y] = NO_DIR;
			}

		frontierNodeList = new LinkedList();
		frontierNodeList.add(new FrontierNode(myX, myY, 0, NO_DIR));

		do {
			final LinkedList newFrontierNodeList = new LinkedList();

			final Iterator it = frontierNodeList.iterator();
			while(it.hasNext()) {
				final FrontierNode fn = (FrontierNode)it.next();

				if (passableMap[fn.x][fn.y]
					&& (fn.dist < distMap[fn.x][fn.y])) {

					distMap[fn.x][fn.y] = fn.dist;
					dirMap[fn.x][fn.y] = fn.dir;
					for (dir = MIN_DIR; dir <= MAX_DIR; dir++) {
						newFrontierNodeList.add(new FrontierNode(fn, dir));
					}
				}
			}
			frontierNodeList = newFrontierNodeList;
			
		} while (!frontierNodeList.isEmpty());
	}

	private static final class FrontierNode {
		public final int x;
		public final int y;
		public final int dist;
		public final byte dir; // dir to get to this node in path from start

		public FrontierNode(int x, int y, int dist, byte dir) {
			this.x = x;
			this.y = y;
			this.dist = dist;
			this.dir = dir; 
		}

		public FrontierNode(final FrontierNode oldNode, final byte dir) {
			this.dist = oldNode.dist + 1;
			this.dir = dir;

			switch (dir) {
				case DistMap.NORTH:
					this.x = oldNode.x;
					this.y = oldNode.y + 1;
					break;
				case DistMap.SOUTH:
					this.x = oldNode.x;
					this.y = oldNode.y - 1;
					break;
				case DistMap.EAST:
					this.x = oldNode.x + 1;
					this.y = oldNode.y;
					break;
				case DistMap.WEST:
					this.x = oldNode.x - 1;
					this.y = oldNode.y;
					break;
				default:
					throw new RuntimeException();
			}
		}
	}
}

