/*
 * ICFP 2002 Programming Contest
 * Entry by Seth Bruder
 *
 */

package sethbruder.icfp2002.entry.player;

import java.util.HashSet;
import java.util.Set;
import java.util.HashMap;
import java.util.Map;
import java.util.LinkedList;
import java.util.Iterator;

import sethbruder.icfp2002.entry.*;

final class Facts {

	public final int myId;
	public final Board board;
	public final int capacity;

	// Map from playerIds (Integer) to PlayerState
	public final Map otherPlayerStates;

	public final PlayerState myState;

	// Map from Board.Points to BaseState (includes package drop locations)
	public final Map bases;

	public int cash;
	public int time;
	public int score;

	// LinkedList of Board.Point
	private final LinkedList recentPositions;

	public Facts(InitInfo ii) {
		this.myId = ii.playerId;
		this.board = ii.board;
		this.capacity = ii.capacity;

		this.otherPlayerStates = PlayerState.getInitPlayerStates(ii);
		this.myState = (PlayerState)otherPlayerStates.get(new Integer(myId));
		otherPlayerStates.remove(new Integer(myId));

		if (this.myState == null) throw new IllegalArgumentException();

		this.bases = BaseState.getInitBaseStates(myState, ii);
		this.cash = ii.cash;
		this.time = 0;
		this.score = 0;

		this.recentPositions = new LinkedList();
	}

	public void assertTurnOutcome(Move move, TurnOutcome outcome) {

		this.time++;
		this.cash -= Math.abs(move.bid);

		int oldScore = score;

		PlayerMoveOutcome[] pmos
			= (PlayerMoveOutcome[])outcome.playerMoveOutcomes
			.values().toArray(new PlayerMoveOutcome[0]);

		// First add any new players and update any old ones that still live

		PlayerState ps;

		for (int i = 0; i < pmos.length; i++) {
			if (pmos[i].playerId == myId) {
				myState.update(pmos[i]);
				score += BaseState.update(bases, myState.x, myState.y,
										  pmos[i]);
			}
			else {
				ps = (PlayerState)otherPlayerStates
					.get(new Integer(pmos[i].playerId));
				if (ps == null) {
					ps = new PlayerState(pmos[i]);
					otherPlayerStates.put(new Integer(ps.playerId), ps);
				}
				else {
					ps.update(pmos[i]);
				}
				BaseState.update(bases, ps.x, ps.y, pmos[i]);
			}
		}
		BaseState.update(bases, myState.x, myState.y, outcome.packagesHere,
						 this.time);
		
		// Remove any dead players
		
		PlayerMoveOutcome pmo;

		PlayerState[] pss
			= (PlayerState[])otherPlayerStates
			.values().toArray(new PlayerState[0]);
		for (int i = 0; i < pss.length; i++) {
			pmo = (PlayerMoveOutcome)outcome
				.playerMoveOutcomes.get(new Integer(pss[i].playerId));
			if (pmo == null)
				otherPlayerStates.remove(new Integer(pss[i].playerId));
		}

		if (score != oldScore) recentPositions.clear();
		else {
			recentPositions.addFirst(new Board.Point(myState.x, myState.y));
			while (recentPositions.size() > 16)
				recentPositions.removeLast();
		}
	}

	/**
	 * @return true iff been here four times in the last sixteen turns
	 * wichout a change in score
	 **/
	public boolean getHaveBeenStuckForAWhile() {

		if (recentPositions.size() < 4) return false;
		Board.Point currentPoint = (Board.Point)recentPositions.getFirst();
		Iterator it = recentPositions.listIterator(1);
		int count = 0;
		while(it.hasNext()) {
			Board.Point p = (Board.Point)it.next();
			if (p.equals(currentPoint)) count++;
			if (count >= 3) return true;
		}
		return false;
	}

	public void clearHaveBeenStuckForAWhile() {
		this.recentPositions.clear();
	}

	public String toString() {
		StringBuffer sb = new StringBuffer();

		sb.append("\nPlayerFacts:\n");
		sb.append("------------\n");
		sb.append("Bases:\n");
		BaseState[] bss
			= (BaseState[])bases.values().toArray(new BaseState[0]);
		for (int i = 0; i < bss.length; i++) {
			if (i > 0) sb.append(",\n");
			sb.append("   [" + bss[i].toString() + "]");
		}
		sb.append("\n");

		sb.append("Other Players' State:\n");
		PlayerState[] pss
			= (PlayerState[])otherPlayerStates
			.values().toArray(new PlayerState[0]);
		for (int i = 0; i < pss.length; i++) {
			if (i > 0) sb.append(",\n");
			sb.append("   " + pss[i].toString());
		}
		sb.append("\n");

		sb.append("My State: ");
		sb.append("score=" + score + ", cash=" + cash
				  + ", time=" + time + ", " + myState
				  + ", stuck=" + getHaveBeenStuckForAWhile());
		sb.append("\n" + board.toString(myState.x, myState.y,
										getOtherPlayerPointSet()));
		sb.append("------------\n");

		return sb.toString();
	}

	private Set getOtherPlayerPointSet() {
		Set rv = new HashSet();

		Iterator it = otherPlayerStates.values().iterator();
		while (it.hasNext()) {
			PlayerState ps = (PlayerState)it.next();
			rv.add(new Board.Point(ps.x, ps.y));
		}
		return rv;
	}
}
