/*
 * ICFP 2002 Programming Contest
 * Entry by Seth Bruder
 *
 */

package sethbruder.icfp2002.entry.player;

import java.util.Random;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.Arrays;

import sethbruder.icfp2002.entry.*;

public final class Player {

	private final Facts facts;
	private final Random random;

	private int randomBase;
	private boolean preferToAvoidCollisions;

	public Player(InitInfo initInfo) {
		this.facts = new Facts(initInfo);
		this.random = new Random(1);
		this.randomBase = -1;
		this.preferToAvoidCollisions = false;
	}

	public Move getMove() {

		Move m = null;

		if (facts.getHaveBeenStuckForAWhile()) {
			preferToAvoidCollisions = !preferToAvoidCollisions;
			facts.clearHaveBeenStuckForAWhile();
		}

		if (preferToAvoidCollisions) {
			try {
				m = getMove(true);
			}
			catch(Exception e) {
				Debug.println("Player.getMove(true) yielded exception: " + e);
				m = null;
			}
		}

		if (m == null) {
			m = getMove(false);
		}

		return m;
	}

	public Move getMove(boolean otherPlayersAreUnpassable) {

		final Move m;

		DistMap distMap = new DistMap(facts, otherPlayersAreUnpassable);
		Board.Point p = new Board.Point(facts.myState.x, facts.myState.y);
		BaseState bs;
		BaseState[] bss;
		int dist, shortestDist;
		BaseState nearestBase = null;
		PackageInfo nearestDrop = null;
		LinkedList dropList = new LinkedList();

		if (facts.myState.packages.isEmpty()) {
			bs = (BaseState)facts.bases.get(p);
			if ((bs != null) && !bs.packages.isEmpty()) {
				m = createPickupMove(bs, distMap, otherPlayersAreUnpassable);
			}
			else {
				bss = (BaseState[])facts.bases
					.values().toArray(new BaseState[0]);

				shortestDist = facts.board.getMaxDist();
				for (int i = 0; i < bss.length; i++) {
					dist = distMap.getDist(bss[i].x, bss[i].y);
					if ((dist < shortestDist)
						&& (!bss[i].packages.isEmpty()
							|| (bss[i].timeOfLastVisit <= 0))) {

						shortestDist = dist;
						nearestBase = bss[i];
					}
				}
				if (nearestBase == null) { // all bases empty
					if (otherPlayersAreUnpassable)
						throw new RuntimeException("All reachable bases "
												   + "empty.");

					if ((randomBase < 0) || (randomBase > bss.length)
						|| ((facts.myState.x == bss[randomBase].x)
							&& (facts.myState.y == bss[randomBase].y))) 
						randomBase = random.nextInt(bss.length);
					nearestBase = bss[randomBase];
				}
				int type
					= distMap.getFirstMoveType(nearestBase.x,
											   nearestBase.y);
				if (type == Move.NULL_MOVE_TYPE) {
					for (int i = 0; i < bss.length; i++) {
						nearestBase = bss[i];
						type = distMap.getFirstMoveType(nearestBase.x,
														nearestBase.y);
						if (type != Move.NULL_MOVE_TYPE) break;
					}
					if (type == Move.NULL_MOVE_TYPE) {
						if (otherPlayersAreUnpassable) {
							throw new RuntimeException("No reachable home"
													   + " base.");
						}
						type = Move.DROP;
					}
				}
				m = new Move(-1, type, null);
			}
		}
		else {
			PackageInfo[] pis = (PackageInfo[])facts.myState.packages
				.toArray(new PackageInfo[0]);

			shortestDist = facts.board.getMaxDist();
			for (int i = 0; i < pis.length; i++) {
				dist = distMap.getDist(pis[i].x, pis[i].y);
				if (dist < shortestDist) {
					shortestDist = dist;
					nearestDrop = pis[i];
					dropList = new LinkedList();
					dropList.add(nearestDrop);
				}
				else if ((dist == shortestDist)
						 && (pis[i].x == nearestDrop.x)
						 && (pis[i].y == nearestDrop.y)) { // same drop point

					dropList.add(pis[i]);
				}
			}
			if (nearestDrop == null) {
				Debug.println("Could not find path to any drop points!");
				if (otherPlayersAreUnpassable)
					throw new RuntimeException("No path to drop points");
				m = new Move(1, Move.DROP, new int[] {pis[0].id});
			}
			else {
				if ((nearestDrop.x == p.x) && (nearestDrop.y == p.y)) {
					m = new Move(1, Move.DROP, dropListToIntArray(dropList));
				}
				else {
					int type = distMap.getFirstMoveType(nearestDrop.x,
														nearestDrop.y);
					if (type == Move.NULL_MOVE_TYPE) {
						if (otherPlayersAreUnpassable)
							throw new RuntimeException("No path to drop"
													   + " points");
						m = new Move(1, Move.DROP, 
									 dropListToIntArray(dropList));
					}
					else
						m = new Move(-1, type, null);
				}
			}
		}

		return m;
	}

	public void assertTurnOutcome(Move move, TurnOutcome outcome) {
		facts.assertTurnOutcome(move, outcome);
	}


	public String toString() {
		return facts.toString();
	}


	/**
	 * Assumes that bs has one or more packages.
	 * Assumes that we have no load yet.
	 **/
	private Move createPickupMove(final BaseState bs, final DistMap distMap,
								  boolean otherPlayersAreUnpassable) {

		int i;
		float eval;
		PackageInfo pi;
		int remainingCapacity = facts.capacity;
		LinkedList packageList = new LinkedList();

		final PackageInfo[] pis
			= (PackageInfo[])bs.packages.toArray(new PackageInfo[0]);

		//
		// Find best single package (i.e. if only one package were to be
		// picked up)
		//

		PackageInfo bestPackage = null;
		float bestEval = Float.NEGATIVE_INFINITY;

		for (i = 0; i < pis.length; i++) {
			eval = evalPackage(pis[i], distMap).eval;
			if ((eval > bestEval) && (pis[i].weight <= remainingCapacity)) {
				bestEval = eval;
				bestPackage = pis[i];
			}
		}
		if (bestPackage == null) throw new RuntimeException();

		packageList.add(bestPackage);
		remainingCapacity -= bestPackage.weight;

		//
		// Now evaluate all remaining packages relative to the position of
		// the best package, and take the best (the bestPackage destination
		// is adopted as a delivery hub)
		//

		final DistMap hubDistMap
			= new DistMap(facts, bestPackage.x, bestPackage.y,
						  otherPlayersAreUnpassable);

		final PackageEval[] pes = new PackageEval[pis.length - 1];
		int j = 0;
		for (i = 0; i < pis.length; i++) {
			if (pis[i] != bestPackage)
				pes[j++] = evalPackage(pis[i], hubDistMap);
		}

		Arrays.sort(pes);
		
		for (i = 0; (i < pes.length) && (remainingCapacity > 0); i++) {
			if (pes[i].pi.weight <= remainingCapacity) {
				packageList.add(pes[i].pi);
				remainingCapacity -= pes[i].pi.weight;
			}
		}

		final int[] rv = new int[packageList.size()];
		Iterator it = packageList.iterator();
		i = 0;
		while (it.hasNext()) {
			rv[i++] = ((PackageInfo)it.next()).id;
		}

		return new Move(1, Move.PICK, rv);
	}

	private PackageEval evalPackage(final PackageInfo pi,
									final DistMap distMap) {
		
		float dist = (float)distMap.getDist(pi.x, pi.y);
		float payoff = (float)pi.weight;
		float avgPayoffPerTurn = (facts.score != 0) 
			? (float)facts.score / (float)Math.max(facts.time, 1)
			: 1.0f;

		float avgTurnsPerDropDist = 1.75f;

		float eval = payoff - (dist * avgTurnsPerDropDist * avgPayoffPerTurn);
		return new PackageEval(pi, eval);
	}

	private int[] dropListToIntArray(LinkedList dropList) {
		PackageInfo[] pis
			= (PackageInfo[])dropList.toArray(new PackageInfo[0]);

		int[] rv = new int[pis.length];
		for (int i = 0; i < rv.length; i++) 
			rv[i] = pis[i].id;

		return rv;
	}

	private static final class PackageEval implements Comparable {
		public final PackageInfo pi;
		public final float eval;

		public PackageEval(PackageInfo pi, float eval) {
			this.pi = pi;
			this.eval = eval;
		}
		
		public int compareTo(Object o) {
			float diff = ((PackageEval)o).eval - this.eval;
			if (diff > 0.0f) return 1;
			else if (diff < 0.0f) return -1;
			else return 0;
		}
	}
}

